import re
from typing import Any

import beangulp
import mt940
from beancount.core import amount, data
from beancount.core.number import D

from tariochbctools.importers.general.deduplication import ReferenceDuplicatesComparator


class Importer(beangulp.Importer):
    """An importer for MT940 files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []
        transactions = mt940.parse(filepath)
        for trx in transactions:
            trxdata = trx.data
            ref = trxdata["bank_reference"]
            if ref:
                metakv = {"ref": ref}
            else:
                metakv = None
            meta = data.new_metadata(filepath, 0, metakv)
            if "entry_date" in trxdata:
                date = trxdata["entry_date"]
            else:
                date = trxdata["date"]
            entry = data.Transaction(
                meta,
                date,
                "*",
                self.prepare_payee(trxdata),
                self.prepare_narration(trxdata),
                data.EMPTY_SET,
                data.EMPTY_SET,
                [
                    data.Posting(
                        self._account,
                        amount.Amount(
                            D(trxdata["amount"].amount), trxdata["amount"].currency
                        ),
                        None,
                        None,
                        None,
                        None,
                    ),
                ],
            )
            entries.append(entry)

        return entries

    cmp = ReferenceDuplicatesComparator()

    def prepare_payee(self, trxdata: dict[str, Any]) -> str:
        return ""

    def prepare_narration(self, trxdata: dict[str, Any]) -> str:
        return trxdata["transaction_details"] + " " + trxdata["extra_details"]



import re
from typing import Any

from tariochbctools.importers.general import mt940importer


class ZkbImporter(mt940importer.Importer):
    def prepare_payee(self, trxdata: dict[str, Any]) -> str:
        return ""

    def prepare_narration(self, trxdata: dict[str, Any]) -> str:
        extra = trxdata["extra_details"]
        details = trxdata["transaction_details"]

        extraReplacements = {}
        extraReplacements[r"Einkauf ZKB Maestro[- ]Karte"] = ""
        extraReplacements[r"LSV:.*"] = "LSV"
        extraReplacements[r"Gutschrift:.*"] = "Gutschrift"
        extraReplacements[r"eBanking:.*"] = "eBanking"
        extraReplacements[r"eBanking Mobile:.*"] = "eBanking Mobile"
        extraReplacements[r"E-Rechnung:.*"] = "E-Rechnung"
        extraReplacements[r"Kontouebertrag:.*"] = "Kontouebertrag:"
        extraReplacements[r"\?ZKB:\d+ "] = ""

        detailsReplacements = {}
        detailsReplacements[r"\?ZI:\?9:\d"] = ""
        detailsReplacements[r"\?ZKB:\d+"] = ""
        detailsReplacements[r"Einkauf ZKB Maestro[- ]Karte Nr. \d+,"] = "Maestro"

        for pattern, replacement in extraReplacements.items():
            extra = re.sub(pattern, replacement, extra)

        for pattern, replacement in detailsReplacements.items():
            details = re.sub(pattern, replacement, details)

        if extra:
            narration = extra.strip() + ": " + details.strip()
        else:
            narration = details.strip()

        return narration



from datetime import datetime
from os import environ
from time import sleep
from typing import Optional

from beancount.core.number import D
from beanprice import source
from dateutil import tz
from ibflex import client, parser


class Source(source.Source):
    def get_latest_price(self, ticker: str) -> source.SourcePrice | None:
        token: str = environ["IBKR_TOKEN"]
        queryId: str = environ["IBKR_QUERY_ID"]

        try:
            response = client.download(token, queryId)
        except client.ResponseCodeError as e:
            if e.code == "1018":
                sleep(10)
                response = client.download(token, queryId)
            else:
                raise e

        statement = parser.parse(response)
        for custStatement in statement.FlexStatements:
            for position in custStatement.OpenPositions:
                if position.symbol.rstrip("z") == ticker:
                    price = D(position.markPrice)
                    timezone = tz.gettz("Europe/Zurich")
                    time = datetime.combine(
                        position.reportDate, datetime.min.time()
                    ).astimezone(timezone)

                    return source.SourcePrice(price, time, position.currency)

        return None

    def get_historical_price(
        self, ticker: str, time: datetime
    ) -> Optional[source.SourcePrice]:
        return None



"""A plugin that inserts an additional price to the base rate by applying
fx rate to a price.
"""

from beancount.core import amount, data, prices

__plugins__ = ["generate"]


def generate(entries, options_map, baseCcy):
    errors = []
    priceMap = prices.build_price_map(entries)

    additionalEntries = []
    for entry in entries:
        if (
            isinstance(entry, data.Price)
            and entry.amount.currency != baseCcy
            and entry.currency != baseCcy
        ):
            fxTuple = tuple([entry.amount.currency, baseCcy])
            fxRate = prices.get_price(priceMap, fxTuple, entry.date)
            if fxRate[1] and not _alreadyExistingPrice(
                priceMap, tuple([entry.currency, baseCcy]), entry.date
            ):
                priceInBaseCcy = amount.Amount(entry.amount.number * fxRate[1], baseCcy)

                additionalEntries.append(
                    data.Price(entry.meta, entry.date, entry.currency, priceInBaseCcy)
                )

    entries.extend(additionalEntries)

    return entries, errors


def _alreadyExistingPrice(priceMap, fxTuple, date):
    if fxTuple in priceMap:
        for alreadyExistingPriceDates in priceMap[fxTuple]:
            if date == alreadyExistingPriceDates[0]:
                return True

    return False



import re
from typing import Any

from tariochbctools.importers.general import mt940importer


class RaiffeisenCHImporter(mt940importer.Importer):
    """An importer for MT940 from Raiffeisen CH"""

    """To get the correct file, choose SWIFT -> 'Période prédéfinie du relevé de compte' -> Sans détails"""

    def prepare_payee(self, trxdata: dict[str, Any]) -> str:
        return ""

    def prepare_narration(self, trxdata: dict[str, Any]) -> str:
        extra = trxdata["extra_details"]
        details = trxdata["transaction_details"]

        detailsReplacements = {}
        detailsReplacements[r"\n"] = ", "

        for pattern, replacement in detailsReplacements.items():
            details = re.sub(pattern, replacement, details)

        if extra:
            narration = extra.strip() + ": " + details.strip()
        else:
            narration = details.strip()

        return narration



import csv
import logging
import re
from datetime import timedelta
from typing import Any

import beangulp
from beancount.core import amount, data
from beancount.core.number import ZERO, D
from dateutil.parser import parse


class Importer(beangulp.Importer):
    """An importer for Revolut CSV files."""

    def __init__(
        self, filepattern: str, account: data.Account, currency: str, fee: Any = None
    ):
        self._filepattern = filepattern
        self._account = account
        self.currency = currency
        self._fee = fee

    def name(self) -> str:
        return super().name() + self._account

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        with open(filepath) as csvfile:
            reader = csv.DictReader(
                csvfile,
                [
                    "Type",
                    "Product",
                    "Started Date",
                    "Completed Date",
                    "Description",
                    "Amount",
                    "Fee",
                    "Currency",
                    "State",
                    "Balance",
                ],
                delimiter=",",
                skipinitialspace=True,
            )
            next(reader)
            is_fee_mode = self._fee is not None
            for row in reader:
                try:
                    bal = D(row["Balance"].replace("'", "").strip())
                    amount_raw = D(row["Amount"].replace("'", "").strip())
                    amt = amount.Amount(amount_raw, row["Currency"])
                    balance = amount.Amount(bal, self.currency)
                    book_date = parse(row["Completed Date"].strip()).date()
                    fee_amt_raw = D(row["Fee"].replace("'", "").strip())
                    fee = amount.Amount(-fee_amt_raw, row["Currency"])
                except Exception as e:
                    logging.warning(e)
                    continue

                if is_fee_mode and fee_amt_raw == ZERO:
                    continue

                postings = [
                    data.Posting(self._account, amt, None, None, None, None),
                ]
                description = row["Description"].strip()
                if is_fee_mode:
                    postings = [
                        data.Posting(self._account, fee, None, None, None, None),
                        data.Posting(
                            self._fee["account"], -fee, None, None, None, None
                        ),
                    ]
                    description = f"Fees for {description}"

                assert isinstance(
                    description, str
                ), "Actual type of description is " + str(type(description))

                entry = data.Transaction(
                    data.new_metadata(filepath, 0, {}),
                    book_date,
                    "*",
                    "",
                    description,
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    postings,
                )
                entries.append(entry)

            if not is_fee_mode:
                # only add balance after the last (newest) transaction
                try:
                    book_date = book_date + timedelta(days=1)
                    entry = data.Balance(
                        data.new_metadata(filepath, 0, {}),
                        book_date,
                        self._account,
                        balance,
                        None,
                        None,
                    )
                    entries.append(entry)
                except NameError:
                    pass

        return entries



import csv
import re

import beangulp
from beancount.core import amount, data
from beancount.core.number import D
from dateutil.parser import parse


class Importer(beangulp.Importer):
    """An importer for Neon CSV files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def name(self) -> str:
        return super().name() + self._account

    def account(self, filepath: str) -> data.Account:
        return self._account

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        with open(filepath) as csvfile:
            reader = csv.DictReader(
                csvfile,
                [
                    "Date",
                    "Amount",
                    "Original amount",
                    "Original currency",
                    "Exchange rate",
                    "Description",
                    "Subject",
                    "Category",
                    "Tags",
                    "Wise",
                    "Spaces",
                ],
                delimiter=";",
                skipinitialspace=True,
            )
            next(reader)
            for row in reader:
                book_date = parse(row["Date"].strip()).date()
                amt = amount.Amount(D(row["Amount"]), "CHF")
                metakv = {
                    "category": row["Category"],
                }
                if row["Original currency"] != "":
                    metakv["original_currency"] = row["Original currency"]
                    metakv["original_amount"] = row["Original amount"]
                    metakv["exchange_rate"] = row["Exchange rate"]

                meta = data.new_metadata(filepath, 0, metakv)
                description = row["Description"].strip()
                if row["Subject"].strip() != "":
                    description = description + ": " + row["Subject"].strip()

                entry = data.Transaction(
                    meta,
                    book_date,
                    "*",
                    "",
                    description,
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    [
                        data.Posting(self._account, amt, None, None, None, None),
                    ],
                )
                entries.append(entry)

        return entries



import re
from datetime import datetime

import beangulp
import camelot
from beancount.core import amount, data
from beancount.core.number import D


class Importer(beangulp.Importer):
    """An importer for Viseca One Card Statement PDF files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account
        self.currency = "CHF"

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def createEntry(
        self, filepath: str, date: str, entryAmount: str | None, text: str
    ) -> data.Transaction:
        amt = None
        if not entryAmount:
            entryAmount = ""
        entryAmount = entryAmount.replace("'", "")
        if "-" in entryAmount:
            amt = amount.Amount(D(entryAmount.strip(" -")), "CHF")
        else:
            amt = -amount.Amount(D(entryAmount), "CHF")

        book_date = datetime.strptime(date, "%d.%m.%y").date()

        meta = data.new_metadata(filepath, 0)
        return data.Transaction(
            meta,
            book_date,
            "*",
            "",
            text.strip(),
            data.EMPTY_SET,
            data.EMPTY_SET,
            [
                data.Posting(self._account, amt, None, None, None, None),
            ],
        )

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        p = re.compile(r"^\d\d\.\d\d\.\d\d$")

        columns = ["100,132,400,472,523"]

        firstPageTables = camelot.read_pdf(
            filepath,
            flavor="stream",
            pages="1",
            table_areas=["65,435,585,100"],
            columns=columns,
            split_text=True,
        )
        otherPageTables = camelot.read_pdf(
            filepath,
            flavor="stream",
            pages="2-end",
            table_areas=["65,670,585,100"],
            columns=columns,
            split_text=True,
        )

        tables = [*firstPageTables, *otherPageTables]

        for table in tables:
            df = table.df

            # skip incompatible tables
            if df.columns.size != 6:
                continue

            lastTrxDate = None
            lastAmount = None
            lastDetails = ""
            for _, row in df.iterrows():
                date, valueDate, details, _, _, amountChf = tuple(row)

                if date and not p.match(date) or "Totalbetrag" in details:
                    continue

                trxDate = valueDate
                details = details.strip()

                if amountChf:
                    if lastTrxDate:
                        entries.append(
                            self.createEntry(
                                filepath, lastTrxDate, lastAmount, lastDetails
                            )
                        )

                    lastTrxDate = trxDate
                    lastAmount = amountChf
                    lastDetails = ""

                lastDetails += details + " "

            if lastTrxDate:
                entries.append(
                    self.createEntry(filepath, lastTrxDate, lastAmount, lastDetails)
                )

        return entries



import csv
import re
from collections.abc import Iterable
from datetime import date

import beangulp
from beancount.core import amount, data
from beancount.core.number import D
from beancount.core.position import CostSpec
from dateutil.parser import parse

from tariochbctools.importers.general.priceLookup import PriceLookup


class Importer(beangulp.Importer):
    """An importer for Fidelity Netbenefits Activity CSV files."""

    def __init__(
        self,
        filepattern: str,
        cashAccount: str,
        investmentAccount: str,
        dividendAccount: str,
        taxAccount: str,
        capGainAccount: str,
        symbol: str,
        ignoreTypes: Iterable[str],
        baseCcy: str,
    ):
        self.filepattern = filepattern
        self.cashAccount = cashAccount
        self.investmentAccount = investmentAccount
        self.dividendAccount = dividendAccount
        self.taxAccount = taxAccount
        self.capGainAccount = capGainAccount
        self.symbol = symbol
        self.ignoreTypes = ignoreTypes
        self.baseCcy = baseCcy

    def name(self) -> str:
        return super().name() + self.cashAccount

    def identify(self, filepath: str) -> bool:
        return re.search(self.filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self.cashAccount

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        self.priceLookup = PriceLookup(existing, self.baseCcy)

        with open(filepath) as csvfile:
            reader = csv.DictReader(
                csvfile,
                [
                    "Transaction date",
                    "Transaction type",
                    "Investment name",
                    "Shares",
                    "Amount",
                ],
                delimiter=",",
                skipinitialspace=True,
            )
            next(reader)
            for row in reader:
                if not row["Transaction type"]:
                    break

                if row["Transaction type"] in self.ignoreTypes:
                    continue

                book_date = parse(row["Transaction date"].strip()).date()
                amt = amount.Amount(D(row["Amount"].replace("$", "")), "USD")
                shares = None
                if row["Shares"] != "-":
                    shares = amount.Amount(D(row["Shares"]), self.symbol)

                metakv: data.Meta = {}

                if not amt and not shares:
                    continue

                meta = data.new_metadata(filepath, 0, metakv)
                description = row["Transaction type"].strip()

                if "TAX" in description:
                    postings = self.__createDividend(amt, book_date, self.taxAccount)
                elif "DIVIDEND" in description:
                    postings = self.__createDividend(
                        amt, book_date, self.dividendAccount
                    )
                elif "YOU BOUGHT" in description:
                    postings = self.__createBuy(amt, shares, book_date)
                elif "YOU SOLD" in description:
                    postings = self.__createSell(amt, shares, book_date)
                else:
                    postings = [
                        data.Posting(self.cashAccount, amt, None, None, None, None),
                    ]

                    if shares is not None:
                        postings.append(
                            data.Posting(
                                self.investmentAccount, shares, None, None, None, None
                            ),
                        )

                entry = data.Transaction(
                    meta,
                    book_date,
                    "*",
                    "",
                    description,
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    postings,
                )
                entries.append(entry)

        return entries

    def __createBuy(
        self, amt: amount, shares: amount, book_date: date
    ) -> list[data.Posting]:
        price = self.priceLookup.fetchPrice("USD", book_date)
        cost = CostSpec(
            number_per=None,
            number_total=round(-amt.number * price.number, 2),
            currency=self.baseCcy,
            date=None,
            label=None,
            merge=None,
        )
        postings = [
            data.Posting(self.investmentAccount, shares, cost, None, None, None),
            data.Posting(self.cashAccount, amt, None, price, None, None),
        ]

        return postings

    def __createSell(
        self, amt: amount, shares: amount, book_date: date
    ) -> list[data.Posting]:
        price = self.priceLookup.fetchPrice("USD", book_date)
        cost = CostSpec(
            number_per=None,
            number_total=None,
            currency=None,
            date=None,
            label=None,
            merge=None,
        )
        postings = [
            data.Posting(self.investmentAccount, shares, cost, None, None, None),
            data.Posting(self.cashAccount, amt, None, price, None, None),
            data.Posting(self.capGainAccount, None, None, None, None, None),
        ]

        return postings

    def __createDividend(
        self, amt: amount, book_date: date, incomeAccount: str
    ) -> list[data.Posting]:
        price = self.priceLookup.fetchPrice("USD", book_date)
        postings = [
            data.Posting(
                self.investmentAccount,
                amount.Amount(D(0), self.symbol),
                None,
                None,
                None,
                None,
            ),
            data.Posting(self.cashAccount, amt, None, price, None, None),
            data.Posting(incomeAccount, None, None, None, None, None),
        ]

        return postings



from datetime import date
from os import path

import beangulp
import requests
import yaml
from beancount.core import amount, data
from beancount.core.number import D

from tariochbctools.importers.general.deduplication import ReferenceDuplicatesComparator


class HttpServiceException(Exception):
    pass


class Importer(beangulp.Importer):
    """An importer for Nordigen API (e.g. for Revolut)."""

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("nordigen.yaml")

    def account(self, filepath: str) -> data.Entries:
        return ""

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)

        r = requests.post(
            "https://ob.nordigen.com/api/v2/token/new/",
            data={
                "secret_id": config["secret_id"],
                "secret_key": config["secret_key"],
            },
        )
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise HttpServiceException(e, e.response.text)

        token = r.json()["access"]
        headers = {"Authorization": "Bearer " + token}

        entries = []
        for account in config["accounts"]:
            accountId = account["id"]
            assetAccount = account["asset_account"]
            r = requests.get(
                f"https://ob.nordigen.com/api/v2/accounts/{accountId}/transactions/",
                headers=headers,
            )
            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise HttpServiceException(e, e.response.text)

            transactions = sorted(
                r.json()["transactions"]["booked"], key=lambda trx: trx["bookingDate"]
            )
            for trx in transactions:
                if "transactionId" in trx:
                    metakv = {
                        "nordref": trx["transactionId"],
                    }
                else:
                    metakv = {}

                if "creditorName" in trx:
                    metakv["creditorName"] = trx["creditorName"]
                if "debtorName" in trx:
                    metakv["debtorName"] = trx["debtorName"]
                if "currencyExchange" in trx:
                    instructedAmount = trx["currencyExchange"]["instructedAmount"]
                    metakv["original"] = (
                        instructedAmount["currency"] + " " + instructedAmount["amount"]
                    )
                meta = data.new_metadata("", 0, metakv)
                trxDate = date.fromisoformat(trx["bookingDate"])
                narration = ""
                if "remittanceInformationUnstructured" in trx:
                    narration += trx["remittanceInformationUnstructured"]
                if "remittanceInformationUnstructuredArray" in trx:
                    narration += " ".join(trx["remittanceInformationUnstructuredArray"])
                entry = data.Transaction(
                    meta,
                    trxDate,
                    "*",
                    "",
                    narration,
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    [
                        data.Posting(
                            assetAccount,
                            amount.Amount(
                                D(str(trx["transactionAmount"]["amount"])),
                                trx["transactionAmount"]["currency"],
                            ),
                            None,
                            None,
                            None,
                            None,
                        ),
                    ],
                )
                entries.append(entry)

        return entries

    cmp = ReferenceDuplicatesComparator(["nordref"])



import re
from datetime import timedelta

import beangulp
import camelot
import pandas as pd
from beancount.core import amount, data
from beancount.core.number import D
from dateutil.parser import parse

from tariochbctools.importers.general.deduplication import ReferenceDuplicatesComparator


class Importer(beangulp.Importer):
    """An importer for Bank Cler ZAK PDF files files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account

    def account(self, filepath: str) -> data.Account:
        return self._account

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def createEntry(
        self, filepath: str, date: str, amt: str, text: str
    ) -> data.Transaction:
        bookingNrRgexp = re.compile(r"BC Buchungsnr. (?P<bookingRef>\d+)$")
        m = bookingNrRgexp.search(text)
        if m:
            bookingRef = m.group("bookingRef")
            text = re.sub(bookingNrRgexp, "", text)

        meta = data.new_metadata(filepath, 0, {"zakref": bookingRef})
        return data.Transaction(
            meta,
            parse(date.strip(), dayfirst=True).date(),
            "*",
            "",
            text.strip(),
            data.EMPTY_SET,
            data.EMPTY_SET,
            [
                data.Posting(
                    self._account, amount.Amount(D(amt), "CHF"), None, None, None, None
                ),
            ],
        )

    def createBalanceEntry(self, filepath: str, date: str, amt: str) -> data.Balance:
        meta = data.new_metadata(filepath, 0)
        return data.Balance(
            meta,
            parse(date.strip(), dayfirst=True).date() + timedelta(days=1),
            self._account,
            amount.Amount(D(amt), "CHF"),
            None,
            None,
        )

    def cleanNumber(self, number: str | data.Decimal) -> data.Decimal:
        if isinstance(number, str):
            return D(number.replace("'", ""))
        else:
            return number

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        firstPageTables = camelot.read_pdf(
            filepath, flavor="stream", pages="1", table_regions=["60,450,600,170"]
        )
        otherPageTables = camelot.read_pdf(
            filepath, flavor="stream", pages="2-end", table_regions=["60,630,600,170"]
        )

        tables = [*firstPageTables, *otherPageTables]

        df: pd.DataFrame | None = None
        for table in tables:
            cur_df = table.df
            new_header = cur_df.iloc[0]
            cur_df = cur_df[1:]
            cur_df.columns = new_header

            if df is None:
                df = cur_df
            else:
                df = pd.concat([df, cur_df])

        date = None
        text = ""
        amount = None
        saldo = None
        if df:
            for row in df.itertuples():
                if row.Saldo:
                    if date and amount:
                        entries.append(self.createEntry(filepath, date, amount, text))

                    date = None
                    amount = None
                    text = ""

                if row.Valuta:
                    date = row.Valuta

                if row.Text:
                    text += " " + row.Text

                if row.Belastung:
                    amount = -self.cleanNumber(row.Belastung)

                if row.Gutschrift:
                    amount = self.cleanNumber(row.Gutschrift)

                if row.Saldo:
                    saldo = self.cleanNumber(row.Saldo)

            if date and amount:
                entries.append(self.createEntry(filepath, date, amount, text))

            dateRegexp = re.compile(r"\d\d\.\d\d\.\d\d\d\d")
            m = dateRegexp.search(text)
            if m and saldo:
                date = m.group()
                entries.append(self.createBalanceEntry(filepath, date, saldo))

        return entries

    cmp = ReferenceDuplicatesComparator(["zakref"])



import logging
from datetime import timedelta
from os import path
from typing import Any

import beangulp
import dateutil.parser
import requests
import yaml
from beancount.core import amount, data
from beancount.core.number import D

# https://docs.truelayer.com/#retrieve-account-transactions

TX_MANDATORY_ID_FIELDS = ("transaction_id",)

TX_OPTIONAL_ID_FIELDS = (
    "normalised_provider_transaction_id",
    "provider_transaction_id",
)

TX_OPTIONAL_META_ID_FIELDS = (
    "provider_id",
    "provider_reference",
)


class Importer(beangulp.Importer):
    """An importer for Truelayer API (e.g. for Revolut)."""

    def __init__(self):
        self.config = None
        self.clientId = None
        self.clientSecret = None
        self.refreshToken = None
        self.sandbox = None
        self.existing = None
        self.domain = "truelayer.com"

    def _configure(self, filepath: str, existing: data.Entries) -> None:
        with open(filepath, "r") as f:
            self.config = yaml.safe_load(f)
        self.clientId = self.config["client_id"]
        self.clientSecret = self.config["client_secret"]
        self.refreshToken = self.config["refresh_token"]
        self.sandbox = self.clientId.startswith("sandbox")
        self.existing = existing

        if self.sandbox:
            self.domain = "truelayer-sandbox.com"

        if "account" not in self.config and "accounts" not in self.config:
            raise KeyError("At least one of `account` or `accounts` must be specified")

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("truelayer.yaml")

    def account(self, filepath: str) -> data.Account:
        return ""

    def extract(self, filepath: str, existing: data.Entries = None) -> data.Entries:
        self._configure(filepath, existing)

        r = requests.post(
            f"https://auth.{self.domain}/connect/token",
            data={
                "grant_type": "refresh_token",
                "client_id": self.clientId,
                "client_secret": self.clientSecret,
                "refresh_token": self.refreshToken,
            },
        )
        tokens = r.json()
        accessToken = tokens["access_token"]
        headers = {"Authorization": "Bearer " + accessToken}

        entries = []
        entries.extend(self._extract_endpoint_transactions("accounts", headers))
        entries.extend(
            self._extract_endpoint_transactions("cards", headers, invert_sign=True)
        )

        return entries

    def _get_account_for_account_id(self, account_id: str) -> data.Account:
        """
        Find a matching account for the account ID.
        If the user hasn't specified any in the config, return
        the base account.

        Otherwise return None.
        """
        if "accounts" not in self.config:
            return self.config["account"]

        # Empty `accounts` will generate warnings for all accounts
        #  including their account IDs
        if self.config["accounts"] is None:
            return None

        return self.config["accounts"].get(account_id, None)

    def _extract_endpoint_transactions(
        self, endpoint: str, headers: dict[str, str], invert_sign: bool = False
    ) -> data.Entries:
        entries = []
        r = requests.get(
            f"https://api.{self.domain}/data/v1/{endpoint}", headers=headers
        )

        if not r:
            try:
                r.raise_for_status()
            except requests.HTTPError as e:
                logging.warning(e)

            return []

        for account in r.json()["results"]:
            accountId = account["account_id"]

            local_account = self._get_account_for_account_id(accountId)

            if not local_account:
                logging.warning("Ignoring account ID %s", accountId)
                continue

            r = requests.get(
                f"https://api.{self.domain}/data/v1/{endpoint}/{accountId}/transactions",
                headers=headers,
            )
            transactions = sorted(r.json()["results"], key=lambda trx: trx["timestamp"])

            for trx in transactions:
                entries.extend(
                    self._extract_transaction(
                        trx, local_account, transactions, invert_sign
                    )
                )

        return entries

    def _extract_transaction(
        self,
        trx: dict[str, Any],
        local_account: data.Account,
        transactions: list[Any],
        invert_sign: bool,
    ) -> data.Transaction:
        entries = []
        metakv = {}

        id_meta_kvs = {
            k: trx["meta"][k] for k in TX_OPTIONAL_META_ID_FIELDS if trx["meta"].get(k)
        }
        metakv.update(id_meta_kvs)

        id_kvs = {
            k: trx[k]
            for k in TX_MANDATORY_ID_FIELDS + TX_OPTIONAL_ID_FIELDS
            if trx.get(k)
        }
        metakv.update(id_kvs)

        if trx["transaction_classification"]:
            metakv["category"] = trx["transaction_classification"][0]

        meta = data.new_metadata("", 0, metakv)
        trxDate = dateutil.parser.parse(trx["timestamp"]).date()

        tx_amount = D(str(trx["amount"]))
        # avoid pylint invalid-unary-operand-type
        signed_amount = -1 * tx_amount if invert_sign else tx_amount

        entry = data.Transaction(
            meta,
            trxDate,
            "*",
            "",
            trx["description"],
            data.EMPTY_SET,
            data.EMPTY_SET,
            [
                data.Posting(
                    local_account,
                    amount.Amount(signed_amount, trx["currency"]),
                    None,
                    None,
                    None,
                    None,
                ),
            ],
        )
        entries.append(entry)

        if trx["transaction_id"] == transactions[-1]["transaction_id"]:
            balDate = trxDate + timedelta(days=1)
            metakv = {}
            if self.existing is not None:
                for exEntry in self.existing:
                    if (
                        isinstance(exEntry, data.Balance)
                        and exEntry.date == balDate
                        and exEntry.account == local_account
                    ):
                        metakv["__duplicate__"] = True

            meta = data.new_metadata("", 0, metakv)

            # Only if the 'balance' permission is present
            if "running_balance" in trx:
                tx_balance = D(str(trx["running_balance"]["amount"]))
                # avoid pylint invalid-unary-operand-type
                signed_balance = -1 * tx_balance if invert_sign else tx_balance

                entries.append(
                    data.Balance(
                        meta,
                        balDate,
                        local_account,
                        amount.Amount(
                            signed_balance, trx["running_balance"]["currency"]
                        ),
                        None,
                        None,
                    )
                )

        return entries



import argparse
import sys
from typing import Any

import requests


def build_header(token: str) -> dict[str, str]:
    return {"Authorization": "Bearer " + token}


def check_result(result: requests.Response) -> None:
    try:
        result.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(e, e.response.text)


def get_token(secret_id: str, secret_key: str) -> str:
    r = requests.post(
        "https://bankaccountdata.gocardless.com/api/v2/token/new/",
        data={
            "secret_id": secret_id,
            "secret_key": secret_key,
        },
    )
    check_result(r)

    return r.json()["access"]


def list_bank(token: str, country: str) -> None:
    r = requests.get(
        "https://bankaccountdata.gocardless.com/api/v2/institutions/",
        params={"country": country},
        headers=build_header(token),
    )
    check_result(r)

    for asp in r.json():
        print(asp["name"] + ": " + asp["id"])  # noqa: T201


def create_link(token: str, reference: str, bank: str) -> None:
    if not bank:
        raise Exception("Please specify --bank it is required for create_link")
    requisitionId = _find_requisition_id(token, reference)
    if requisitionId:
        print(f"Link for for reference {reference} already exists.")  # noqa: T201
    else:
        r = requests.post(
            "https://bankaccountdata.gocardless.com/api/v2/requisitions/",
            data={
                "redirect": "http://localhost",
                "institution_id": bank,
                "reference": reference,
            },
            headers=build_header(token),
        )
        check_result(r)
        link = r.json()["link"]
        print(f"Go to {link} for connecting to your bank.")  # noqa: T201


def list_accounts(token: str) -> None:
    headers = build_header(token)
    r = requests.get(
        "https://bankaccountdata.gocardless.com/api/v2/requisitions/", headers=headers
    )
    print(r.json())  # noqa: T201
    check_result(r)
    for req in r.json()["results"]:
        reference = req["reference"]
        print(f"Reference: {reference}")  # noqa: T201
        for account in req["accounts"]:
            ra = requests.get(
                f"https://bankaccountdata.gocardless.com/api/v2/accounts/{account}",
                headers=headers,
            )
            check_result(ra)
            acc = ra.json()
            asp = acc["institution_id"]
            iban = acc["iban"]

            ra = requests.get(
                f"https://bankaccountdata.gocardless.com/api/v2/accounts/{account}/details",
                headers=headers,
            )
            check_result(ra)
            accDetails = ra.json()["account"]

            currency = accDetails["currency"]
            owner = accDetails["ownerName"] if "ownerName" in accDetails else "-"
            print(f"{account}: {asp} {owner} {iban} {currency}")  # noqa: T201


def delete_link(token: str, reference: str) -> None:
    requisitionId = _find_requisition_id(token, reference)
    if requisitionId:
        r = requests.delete(
            f"https://bankaccountdata.gocardless.com/api/v2/requisitions/{requisitionId}",
            headers=build_header(token),
        )
        check_result(r)


def _find_requisition_id(token: str, userId: str) -> str | None:
    headers = build_header(token)
    r = requests.get(
        "https://bankaccountdata.gocardless.com/api/v2/requisitions/", headers=headers
    )
    check_result(r)
    for req in r.json()["results"]:
        if req["reference"] == userId:
            return req["id"]

    return None


def parse_args(args: Any) -> Any:
    parser = argparse.ArgumentParser(description="nordigen-config")
    parser.add_argument(
        "--secret_id",
        required=True,
        help="API secret id, can be generated on Nordigen website",
    )
    parser.add_argument(
        "--secret_key",
        required=True,
        help="API secret key, can be generated on Nordigen website",
    )
    parser.add_argument(
        "--country",
        default="GB",
        help="Country Code for list_bank",
    )
    parser.add_argument(
        "--reference",
        default="beancount",
        help="reference for add_bank and delete_bank, needs to be unique",
    )
    parser.add_argument(
        "--bank",
        help="Bank to connect to, see list_banks",
    )
    parser.add_argument(
        "mode",
        choices=[
            "list_banks",
            "create_link",
            "list_accounts",
            "delete_link",
        ],
    )
    return parser.parse_args(args)


def main(args: Any) -> None:
    args = parse_args(args)

    token = get_token(args.secret_id, args.secret_key)

    if args.mode == "list_banks":
        list_bank(token, args.country)
    elif args.mode == "create_link":
        create_link(token, args.reference, args.bank)
    elif args.mode == "list_accounts":
        list_accounts(token)
    elif args.mode == "delete_link":
        delete_link(token, args.reference)


def run() -> None:
    """Entry point for console_scripts"""
    main(sys.argv[1:])



import csv
import re

from beancount.core import amount, data
from beancount.core.number import D
from beangulp import Importer
from dateutil.parser import parse


class SwisscardImporter(Importer):
    """An importer for Swisscard's cashback CSV files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account

    def name(self) -> str:
        return super().name() + self._account

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []
        with open(filepath) as csvfile:
            reader = csv.DictReader(
                csvfile,
                delimiter=",",
                skipinitialspace=True,
            )
            for row in reader:
                book_date = parse(row["Transaction date"].strip(), dayfirst=True).date()
                amt = amount.Amount(-D(row["Amount"]), row["Currency"])
                metakv = {
                    "merchant": row["Merchant Category"],
                    "category": row["Registered Category"],
                }
                meta = data.new_metadata(filepath, 0, metakv)
                description = row["Description"].strip()
                entry = data.Transaction(
                    meta,
                    book_date,
                    "*",
                    "",
                    description,
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    [
                        data.Posting(self._account, amt, None, None, None, None),
                    ],
                )
                entries.append(entry)

        return entries



import datetime
import re
from datetime import timedelta

import beangulp
import camelot
from beancount.core import amount, data
from beancount.core.number import D


class Importer(beangulp.Importer):
    """An importer for Cembra Card Statement PDF files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account
        self.currency = "CHF"

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def createEntry(
        self, filepath: str, date: datetime.date, amt: data.Decimal, text: str
    ) -> data.Transaction:
        meta = data.new_metadata(filepath, 0)
        return data.Transaction(
            meta,
            date,
            "*",
            "",
            text,
            data.EMPTY_SET,
            data.EMPTY_SET,
            [
                data.Posting(self._account, amt, None, None, None, None),
            ],
        )

    def createBalanceEntry(
        self, filepath: str, date: datetime.date, amt: data.Decimal
    ) -> data.Balance:
        meta = data.new_metadata(filepath, 0)
        return data.Balance(meta, date, self._account, amt, None, None)

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        tables = camelot.read_pdf(
            filepath, pages="2-end", flavor="stream", table_areas=["50,700,560,50"]
        )
        for table in tables:
            df = table.df

            # skip incompatible tables
            if df.columns.size != 5:
                continue

            for index, row in df.iterrows():
                trx_date, book_date, text, credit, debit = tuple(row)
                trx_date, book_date, text, credit, debit = (
                    trx_date.strip(),
                    book_date.strip(),
                    text.strip(),
                    credit.strip(),
                    debit.strip(),
                )

                # Transaction entry
                try:
                    book_date = datetime.datetime.strptime(book_date, "%d.%m.%Y").date()
                except Exception:
                    book_date = None

                if book_date:
                    amount = self.getAmount(debit, credit)

                    if amount:
                        entries.append(
                            self.createEntry(filepath, book_date, amount, text)
                        )
                    continue

                # Balance entry
                try:
                    m = re.search(
                        r"Saldo per (\d\d\.\d\d\.\d\d\d\d) zu unseren Gunsten CHF",
                        text,
                    )
                    if m:
                        book_date = m.group(1)
                        book_date = datetime.datetime.strptime(
                            book_date, "%d.%m.%Y"
                        ).date()
                        # add 1 day: cembra provides balance at EOD, but beancount checks it at SOD
                        book_date = book_date + timedelta(days=1)
                except Exception:
                    book_date = None

                if book_date:
                    amount = self.getAmount(debit, credit)

                    if amount:
                        entries.append(
                            self.createBalanceEntry(filepath, book_date, amount)
                        )

        return entries

    def cleanDecimal(self, formattedNumber: str) -> data.Decimal:
        return D(formattedNumber.replace("'", ""))

    def getAmount(self, debit: str, credit: str) -> data.Amount:
        amt = -self.cleanDecimal(debit) if debit else self.cleanDecimal(credit)
        if amt:
            return amount.Amount(amt, self.currency)



from os import path

import beangulp
import blockcypher
import yaml
from beancount.core import amount, data
from beancount.core.number import D

from tariochbctools.importers.general.deduplication import ReferenceDuplicatesComparator
from tariochbctools.importers.general.priceLookup import PriceLookup


class Importer(beangulp.Importer):
    """An importer for Blockchain data."""

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("blockchain.yaml")

    def account(self, filepath: str) -> data.Entries:
        return ""

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        with open(filepath) as file:
            config = yaml.safe_load(file)
        self.config = config
        baseCcy = config["base_ccy"]
        priceLookup = PriceLookup(existing, baseCcy)

        entries = []
        for address in self.config["addresses"]:
            currency = address["currency"]
            addressDetails = blockcypher.get_address_details(
                address["address"], coin_symbol=currency.lower()
            )
            for trx in addressDetails["txrefs"]:
                metakv = {
                    "ref": trx["tx_hash"],
                }
                meta = data.new_metadata(file.name, 0, metakv)

                date = trx["confirmed"].date()
                price = priceLookup.fetchPriceAmount(currency, date)
                cost = data.CostSpec(price, None, baseCcy, None, None, False)

                outputType = "ether" if currency.lower() == "eth" else "btc"
                amt = blockcypher.from_base_unit(trx["value"], outputType)

                entry = data.Transaction(
                    meta,
                    date,
                    "*",
                    "",
                    address["narration"],
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    [
                        data.Posting(
                            address["asset_account"],
                            amount.Amount(D(str(amt)), currency),
                            cost,
                            None,
                            None,
                            None,
                        ),
                    ],
                )
                entries.append(entry)

        return entries

    cmp = ReferenceDuplicatesComparator()



import os
from io import StringIO

import pytest
from beancount import loader
from beancount.parser import printer


@pytest.mark.parametrize(
    "testCase", ["normal", "missing_fx", "entry_already_exists", "issue122"]
)
def test_data(testCase):
    dataDir = os.path.join(
        os.path.dirname(__file__), "data", "generate_base_ccy_prices"
    )
    inputPath = os.path.join(dataDir, testCase + "_input.beancount")
    expectedPath = os.path.join(dataDir, testCase + "_expected.beancount")

    entries, errors, _ = loader.load_file(inputPath)
    if errors:
        printer.print_errors(errors)
        assert False

    actualStrIo = StringIO()
    printer.print_entries(entries, file=actualStrIo)
    actual = actualStrIo.getvalue()

    if os.path.isfile(expectedPath):
        with open(expectedPath, "r") as expectedFile:
            expected = expectedFile.read()
            assert actual == expected
    else:
        with open(expectedPath, "w") as expectedFile:
            expectedFile.write(actual)



import tempfile
from os import path

import yaml
from beancount.core import data
from beangulp import Importer
from imap_tools import MailBox


class MailAdapterImporter(Importer):
    """An importer adapter that fetches file from mails and then calls another importer."""

    def __init__(self, importers: list[Importer]):
        self.importers = importers

    def identify(self, filepath: str) -> bool:
        return "mail.yaml" == path.basename(filepath)

    def account(self, filepath: str) -> data.Account:
        return ""

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        with open(filepath) as file:
            config = yaml.safe_load(file)

        with MailBox(config["host"]).login(
            config["user"], config["password"], initial_folder=config["folder"]
        ) as mailbox:
            result = []
            for msg in mailbox.fetch():
                processed = False
                for att in msg.attachments:
                    with tempfile.TemporaryDirectory() as tmpdirname:
                        attFileName = path.join(tmpdirname, att.filename)
                        with open(attFileName, "wb") as attFile:
                            attFile.write(att.payload)
                            attFile.flush()

                            for delegate in self.importers:
                                if delegate.identify(attFileName):
                                    newEntries = delegate.extract(attFileName, existing)
                                    result.extend(newEntries)
                                    processed = True

                if processed and "targetFolder" in config:
                    mailbox.move(msg.uid, config["targetFolder"])

        return result



import base64
import json
from datetime import date, datetime, timezone
from os import path
from typing import Any
from urllib.parse import urlencode

import beangulp
import dateutil.parser
import requests
import rsa
import urllib3
import yaml
from beancount.core import amount, data
from beancount.core.number import D
from dateutil.relativedelta import relativedelta

from tariochbctools.importers.general.deduplication import ReferenceDuplicatesComparator

http = urllib3.PoolManager()


class Importer(beangulp.Importer):
    """An importer for Transferwise using the API."""

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("transferwise.yaml")

    def account(self, filepath: str) -> data.Account:
        return ""

    def __init__(self, *args: Any, **kwargs: Any):
        if "profileId" in kwargs:
            self.profileId = kwargs.pop("profileId")
        if "startDate" in kwargs:
            self.startDate = kwargs.pop("startDate")
        else:
            self.startDate = datetime.combine(
                date.today() + relativedelta(months=-3),
                datetime.min.time(),
                timezone.utc,
            ).isoformat()
        if "endDate" in kwargs:
            self.endDate = kwargs.pop("endDate")
        else:
            self.endDate = datetime.combine(
                date.today(), datetime.max.time(), timezone.utc
            ).isoformat()
        super().__init__(*args, **kwargs)

    # Based on the Transferwise official example provided under the
    # MIT license
    def _get_statement(
        self,
        currency: str,
        base_url: str,
        statement_type: str = "FLAT",
    ) -> Any:
        params = urlencode(
            {
                "currency": currency,
                "type": statement_type,
                "intervalStart": self.startDate,
                "intervalEnd": self.endDate,
            }
        )

        url = (
            base_url
            + "/v3/profiles/"
            + str(self.profileId)
            + "/borderless-accounts/"
            + str(self.accountId)
            + "/statement.json?"
            + params
        )

        headers = {
            "Authorization": "Bearer " + self.api_token,
            "User-Agent": "tw-statements-sca",
            "Content-Type": "application/json",
        }
        if hasattr(self, "one_time_token"):
            headers["x-2fa-approval"] = self.one_time_token  # type: ignore
            headers["X-Signature"] = self.signature  # type: ignore

        r = http.request("GET", url, headers=headers, retries=False)

        if r.status == 200 or r.status == 201:
            return json.loads(r.data)
        elif r.status == 403 and r.getheader("x-2fa-approval") is not None:
            self.one_time_token = r.getheader("x-2fa-approval")
            self.signature = self._do_sca_challenge()
            return self._get_statement(
                currency=currency,
                base_url=base_url,
                statement_type=statement_type,
            )
        else:
            raise Exception("Failed to get transactions.")

    def _do_sca_challenge(self):
        # Read the private key file as bytes.
        with open(self.private_key_path, "rb") as f:
            private_key_data = f.read()

        private_key = rsa.PrivateKey.load_pkcs1(private_key_data, "PEM")

        # Use the private key to sign the one-time-token that was returned
        # in the x-2fa-approval header of the HTTP 403.
        signed_token = rsa.sign(
            self.one_time_token.encode("ascii"), private_key, "SHA-256"
        )

        # Encode the signed message as friendly base64 format for HTTP
        # headers.
        signature = base64.b64encode(signed_token).decode("ascii")

        return signature

    def extract(self, filepath, existing):
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)
        self.api_token = config["token"]
        baseAccount = config["baseAccount"]
        self.private_key_path = config["privateKeyPath"]

        headers = {"Authorization": "Bearer " + self.api_token}
        if not self.profileId:
            r = requests.get(
                "https://api.transferwise.com/v1/profiles", headers=headers
            )
            profiles = r.json()
            self.profileId = profiles[0]["id"]

        r = requests.get(
            "https://api.transferwise.com/v1/borderless-accounts",
            params={"profileId": self.profileId},
            headers=headers,
        )
        accounts = r.json()
        self.accountId = accounts[0]["id"]

        entries = []
        base_url = "https://api.transferwise.com"
        for account in accounts[0]["balances"]:
            accountCcy = account["currency"]
            if isinstance(baseAccount, dict):
                account_name = baseAccount[accountCcy]
            else:
                account_name = baseAccount + accountCcy
            transactions = self._get_statement(
                currency=accountCcy,
                base_url=base_url,
                statement_type="FLAT",
            )

            for transaction in transactions["transactions"]:
                metakv = {
                    "ref": transaction["referenceNumber"],
                }
                meta = data.new_metadata("", 0, metakv)
                entry = data.Transaction(
                    meta,
                    dateutil.parser.parse(transaction["date"]).date(),
                    "*",
                    "",
                    transaction["details"]["description"],
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    [
                        data.Posting(
                            account_name,
                            amount.Amount(
                                D(str(transaction["amount"]["value"])),
                                transaction["amount"]["currency"],
                            ),
                            None,
                            None,
                            None,
                            None,
                        ),
                    ],
                )
                entries.append(entry)

        return entries

    cmp = ReferenceDuplicatesComparator()



import logging
import uuid
from datetime import datetime
from hashlib import md5
from os import path
from typing import Dict, List, NamedTuple

import beangulp
import requests
import yaml
from beancount.core import amount, data
from beancount.core.number import D
from undictify import type_checked_constructor


@type_checked_constructor(skip=True, convert=True)
class QuickFileTransaction(NamedTuple):
    """Transaction data from QuickFile transaction API"""

    TransactionDate: str
    Reference: str
    Amount: str  # otherwise tiny conversion losses
    TagStatus: str
    TransactionId: str

    def to_beancount_transaction(self, local_account, currency, invert_sign=False):
        tx_amount = D(self.Amount)
        # avoid pylint invalid-unary-operand-type
        signed_amount = -1 * tx_amount if invert_sign else tx_amount

        metakv = {
            "quickfile_id": self.TransactionId,
        }

        meta = data.new_metadata("", 0, metakv)
        date = datetime.fromisoformat(self.TransactionDate).date()

        entry = data.Transaction(
            meta,
            date,
            "*",
            "",
            self.Reference,
            data.EMPTY_SET,
            data.EMPTY_SET,
            [
                data.Posting(
                    local_account,
                    amount.Amount(signed_amount, currency),
                    None,
                    None,
                    None,
                    None,
                ),
            ],
        )
        return entry


@type_checked_constructor(skip=True, convert=True)
class QuickFileResponseMetaData(NamedTuple):
    RecordsetCount: int
    ReturnCount: int
    BankName: str
    BankType: str
    AccountNo: str
    SortCode: str
    Currency: str
    CurrentBalance: str


@type_checked_constructor(skip=True, convert=True)
class QuickFileBankSearch(NamedTuple):
    MetaData: QuickFileResponseMetaData
    Transactions: Dict[str, List[QuickFileTransaction]]


class QuickFile:
    """Encapsulate QuickFile API protocol and data types"""

    DOMAIN = "quickfile.co.uk"
    API_VERSION_SLUG = "1_2"

    def __init__(self, account_number, api_key, app_id):
        self.account_number = account_number
        self.api_key = api_key
        self.app_id = app_id
        self._update_submission_number()

    @staticmethod
    def auth_md5(account_number, api_key, submission_number):
        md5_str = (account_number + api_key + str(submission_number)).encode("utf-8")

        return md5(md5_str).hexdigest()

    def request_header(self):
        auth_md5 = self.auth_md5(
            self.account_number, self.api_key, self.submission_number
        )

        header = {
            "MessageType": "Request",
            "SubmissionNumber": str(self.submission_number),
            "Authentication": {
                "AccNumber": str(self.account_number),
                "MD5Value": auth_md5,
                "ApplicationID": str(self.app_id),
            },
        }
        return header

    def _update_submission_number(self):
        self.submission_number = uuid.uuid4()

    def _post(self, endpoint, endpoint_data):
        header = self.request_header()
        post_data = {"payload": {"Header": header, "Body": endpoint_data}}

        r = requests.post(
            f"https://api.{self.DOMAIN}/{self.API_VERSION_SLUG}/{endpoint}",
            json=post_data,
        )

        if not r:
            try:
                r.raise_for_status()
            except requests.HTTPError as e:
                logging.warning(e)

        self._update_submission_number()

        return r.json()

    def bank_search(
        self, account_number, transaction_count, from_date=None, to_date=None
    ):
        endpoint_data = {
            "SearchParameters": {
                "ReturnCount": str(transaction_count),
                "Offset": "0",
                "OrderResultsBy": "TransactionDate",
                "OrderDirection": "DESC",
                "NominalCode": str(account_number),
            }
        }
        if from_date:
            endpoint_data["SearchParameters"]["FromDate"] = str(from_date)
        if to_date:
            endpoint_data["SearchParameters"]["ToDate"] = str(to_date)
        response = self._post("bank/search", endpoint_data)
        body = response["Bank_Search"]["Body"]
        return QuickFileBankSearch(**body)


class Importer(beangulp.Importer):
    """An importer for QuickFile"""

    def __init__(self):
        self.quickfile = None
        self.config = None
        self.existing = None

    def _configure(self, filepath, existing):
        with open(filepath, "r") as config_file:
            self.config = yaml.safe_load(config_file)
            self.quickfile = QuickFile(
                account_number=self.config["account_number"],
                api_key=self.config["api_key"],
                app_id=self.config["app_id"],
            )
        self.existing = existing

    def identify(self, filepath):
        return path.basename(filepath) == "quickfile.yaml"

    def account(self, filepath):
        return ""

    def extract(self, filepath, existing=None):
        self._configure(filepath, existing)
        entries = []

        for bank_account in self.config["accounts"].keys():
            entries.extend(self._extract_bank_transactions(bank_account))

        return entries

    def _extract_bank_transactions(self, bank_account, invert_sign=False):
        entries = []
        transaction_count = self.config["transaction_count"]  # [0..200]
        from_date = self.config.get("from_date", None)
        to_date = self.config.get("to_date", None)
        response = self.quickfile.bank_search(
            bank_account, transaction_count, from_date, to_date
        )
        metadata = response.MetaData
        transactions = response.Transactions["Transaction"]
        local_account = self.config["accounts"].get(bank_account)

        for trx in transactions:
            entries.extend(
                self._extract_transaction(
                    trx, local_account, metadata, transactions, invert_sign
                )
            )

        return entries

    def _extract_transaction(
        self, trx, local_account, metadata, transactions, invert_sign
    ):
        entries = []

        entry = trx.to_beancount_transaction(
            local_account, metadata.Currency, invert_sign
        )
        entries.append(entry)

        return entries



import re
from datetime import datetime

import beangulp
import camelot
from beancount.core import amount, data
from beancount.core.number import D


class Importer(beangulp.Importer):
    """An importer for radicant Account Statement PDF files."""

    def __init__(self, filepattern: str, account: data.Account):
        self._filepattern = filepattern
        self._account = account
        self.currency = "CHF"

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def cleanAmount(
        self, debit: str | None, credit: str | None
    ) -> amount.Amount | None:
        if debit:
            return -amount.Amount(D(debit.replace("'", "")), self.currency)
        elif credit:
            return amount.Amount(D(credit.replace("'", "")), self.currency)
        else:
            return None

    def createEntry(
        self,
        filepath: str,
        date: str,
        amt: amount.Amount,
        text: str,
        conversionOriginal: str | None,
        conversionRate: str | None,
    ) -> data.Transaction:

        book_date = datetime.strptime(date, "%d.%m.%y").date()

        if conversionOriginal and conversionRate:
            kv = {"original": conversionOriginal, "rate": conversionRate}
            text = text.replace("Amount: " + conversionOriginal, "")
        else:
            kv = None

        meta = data.new_metadata(filepath, 0, kv)
        return data.Transaction(
            meta,
            book_date,
            "*",
            "",
            text.strip(),
            data.EMPTY_SET,
            data.EMPTY_SET,
            [
                data.Posting(self._account, amt, None, None, None, None),
            ],
        )

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        entries = []

        conversionPattern = re.compile(r"(?P<original>.+) at the rate of (?P<rate>.+)")

        tables = camelot.read_pdf(
            filepath,
            flavor="stream",
            pages="all",
            table_regions=["40,470,580,32"],
            columns=["80,300,370,440,500"],
        )

        for table in tables:
            df = table.df

            lastTrxDate = None
            lastAmount = None
            lastDetails = ""
            beforeStart = True
            conversionOriginal = None
            conversionRate = None
            for _, row in df.iterrows():
                date, text, _, debit, credit, _ = tuple(row)

                # skip stuff before
                if beforeStart and "Date" != date:
                    continue
                elif "Date" == date:
                    beforeStart = False
                    continue

                # skip stuff after
                if "Balance as of" in text:
                    break

                trxDate = date
                details = text.strip()
                amt = self.cleanAmount(debit, credit)

                if amt:
                    if lastTrxDate:
                        entries.append(
                            self.createEntry(
                                filepath,
                                lastTrxDate,
                                lastAmount,
                                lastDetails,
                                conversionOriginal,
                                conversionRate,
                            )
                        )

                    lastTrxDate = trxDate
                    lastAmount = amt
                    lastDetails = ""
                    conversionOriginal = None
                    conversionRate = None

                match = conversionPattern.match(details)
                if match:
                    conversionOriginal = match.group("original")
                    conversionRate = match.group("rate")
                else:
                    lastDetails += details + " "

            if lastTrxDate:
                entries.append(
                    self.createEntry(
                        filepath,
                        lastTrxDate,
                        lastAmount,
                        lastDetails,
                        conversionOriginal,
                        conversionRate,
                    )
                )

        return entries



"""A plugin that verifies that on each transaction, all the "portfolios" have
the same weight.
"""

import collections
from collections import defaultdict
from decimal import Decimal
from math import isclose

from beancount.core import convert, data

__plugins__ = ["check"]

NonZeroWeightPerPortfolio = collections.namedtuple(
    "NonZeroWeightPerPortfolio", "source message entry"
)
DifferentWeightPerPortfolio = collections.namedtuple(
    "DifferentWeightPerPortfolio", "source message entry"
)


def check(entries, options_map):
    errors = []

    for entry in data.filter_txns(entries):
        positivePortfolioSums = defaultdict(Decimal)
        negativePortfolioSums = defaultdict(Decimal)
        for posting in entry.postings:
            if posting.meta and "portfolio_check_weight" in posting.meta:
                weight = Decimal(posting.meta["portfolio_check_weight"])
            else:
                weight = round(convert.get_weight(posting).number, 2)
            account = posting.account
            portfolio = account.split(":")[1]
            if weight > 0:
                positivePortfolioSums[portfolio] += weight
            else:
                negativePortfolioSums[portfolio] += weight

        portfolios = set(
            list(positivePortfolioSums.keys()) + list(negativePortfolioSums.keys())
        )
        weight = None
        for portfolio in portfolios:
            positiveWeight = positivePortfolioSums[portfolio]
            negativeWeight = -negativePortfolioSums[portfolio]
            if not isclose(positiveWeight, negativeWeight, abs_tol=0.05):
                errors.append(
                    NonZeroWeightPerPortfolio(
                        entry.meta,
                        f"Weights for portfolio {portfolio} don't equal zero {positiveWeight} != {-negativeWeight}",
                        entry,
                    )
                )
            if (
                weight
                and weight != positiveWeight
                and "skip_cross_portfolio_check" not in entry.meta
            ):
                errors.append(
                    DifferentWeightPerPortfolio(
                        entry.meta, "Not all portfolios have the same weight", entry
                    )
                )
            weight = positiveWeight

    return entries, errors



import json

import pytest
import yaml
from beancount.core.amount import Decimal as D

from tariochbctools.importers.truelayer import importer as tlimp

# pylint: disable=protected-access

TEST_CONFIG = b"""
    account: DefaultAccount
    client_id: sandbox-random
    client_secret: deadc0de-dead-c0de-dead-c0dedeadc0de
    refresh_token: 98D124C0E677865CB2F7D9DE91DC394CEED31DA3469C681B41FB7831F2F9B089
    access_token: eyJhbGciOiJSUzI1NiIsImtpZsomethingSomethingsomethingDarkSideOEJBNk
    accounts:
      hex-account-id-1: Assets:Other
      hex-account-id-2: Assets:Savings
      hex-account-id-3: Liabilities:Mastercard
      hex-account-id-4: Liabilities:Visa
"""

# Example from the truelayer Mock bank
TEST_TRX = b"""
{
  "timestamp": "2021-06-14T00:00:00Z",
  "description": "MT SECURETRADE LIM",
  "transaction_type": "CREDIT",
  "transaction_category": "PURCHASE",
  "transaction_classification": [],
  "amount": -20.5,
  "currency": "GBP",
  "transaction_id": "mandatory-transaction-id-value",
  "provider_transaction_id": "6906e4ecbf4a9775e3",
  "normalised_provider_transaction_id": "txn-4912996e337e5b5f3",
  "running_balance": {
    "currency": "GBP",
    "amount": 332.94
  },
  "meta": {
    "provider_id": "SOME-PROVIDER-ID",
    "provider_reference": "SOME-PROVIDER-REFERENCE",
    "provider_transaction_category": "DEB"
  }
}
"""

# Example from the truelayer Mock bank
TEST_TRX_WITHOUT_IDS = b"""
{
  "timestamp": "2021-06-14T00:00:00Z",
  "description": "MT SECURETRADE LIM",
  "transaction_type": "CREDIT",
  "transaction_category": "PURCHASE",
  "transaction_classification": [],
  "amount": -20.5,
  "currency": "GBP",
  "transaction_id": "81cd222c77f49eb11b72a82478a07dbd",
  "running_balance": {
    "currency": "GBP",
    "amount": 332.94
  },
  "meta": {
    "provider_transaction_category": "DEB"
  }
}
"""


@pytest.fixture(name="tmp_config")
def tmp_config_fixture(tmp_path):
    config = tmp_path / "truelayer.yaml"
    config.write_bytes(TEST_CONFIG)
    yield config


@pytest.fixture(name="importer")
def truelayer_importer_fixture(tmp_config):
    importer = tlimp.Importer()
    importer._configure(tmp_config, [])
    yield importer


@pytest.fixture(name="importer_factory")
def truelayer_importer_factory(tmp_path):
    """A truelayer importer factory for
    generating an importer with a custom config
    """

    def _importer_with_config(custom_config):
        config = tmp_path / "truelayer.yaml"
        config.write_bytes(custom_config)
        importer = tlimp.Importer()
        importer._configure(config, [])
        return importer

    yield _importer_with_config


@pytest.fixture(name="tmp_trx")
def tmp_trx_fixture():
    yield json.loads(TEST_TRX)


def test_identify(importer, tmp_config):
    assert importer.identify(tmp_config)


def test_extract_transaction_simple(importer, tmp_trx):
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    assert entries[0].postings[0].units.number == D(str(tmp_trx["amount"]))


def test_extract_transaction_with_balance(importer, tmp_trx):
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    # one entry, one balance
    assert len(entries) == 2
    assert entries[1].amount.number == D(str(tmp_trx["running_balance"]["amount"]))


def test_extract_transaction_invert_sign(importer, tmp_trx):
    """Show that sign inversion works"""
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=True
    )
    assert entries[0].postings[0].units.number == -D(str(tmp_trx["amount"]))


@pytest.mark.parametrize("id_field", tlimp.TX_MANDATORY_ID_FIELDS)
def test_extract_transaction_has_transaction_id(importer, tmp_trx, id_field):
    """Ensure mandatory IDs are in extracted transactions."""
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    assert entries[0].meta[id_field] == tmp_trx[id_field]


@pytest.mark.parametrize("id_field", tlimp.TX_OPTIONAL_ID_FIELDS)
def test_trx_id(importer, tmp_trx, id_field):
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    assert entries[0].meta[id_field] == tmp_trx[id_field]


@pytest.mark.parametrize("id_field", tlimp.TX_OPTIONAL_ID_FIELDS)
def test_trx_id_is_optional(importer, id_field):
    tmp_trx = json.loads(TEST_TRX_WITHOUT_IDS)
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    assert entries[0].meta.get(id_field) is None


@pytest.mark.parametrize("id_field", tlimp.TX_OPTIONAL_META_ID_FIELDS)
def test_trx_meta_id(importer, tmp_trx, id_field):
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    assert entries[0].meta[id_field] == tmp_trx["meta"][id_field]


@pytest.mark.parametrize("id_field", tlimp.TX_OPTIONAL_META_ID_FIELDS)
def test_trx_meta_id_is_optional(importer, id_field):
    tmp_trx = json.loads(TEST_TRX_WITHOUT_IDS)
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", [tmp_trx], invert_sign=False
    )
    assert entries[0].meta.get(id_field) is None


@pytest.mark.parametrize("account_id", yaml.safe_load(TEST_CONFIG)["accounts"].keys())
def test_get_account_for_account_id(importer, account_id):
    assert (
        importer._get_account_for_account_id(account_id)
        == importer.config["accounts"][account_id]
    )


def test_get_account_for_account_id_returns_none(importer):
    assert importer._get_account_for_account_id("unknown-account-id") is None


def test_accounts_config_is_optional(importer_factory):
    TEST_CONFIG_WITHOUT_ACCOUNTS = b"""
        account: DefaultAccount
        client_id: sandbox-random
        client_secret: deadc0de-dead-c0de-dead-c0dedeadc0de
        refresh_token: 98D124C0E677865CB2F7D9DE91DC394CEED31DA3469C681B41FB7831F2F9B089
        access_token: eyJhbGciOiJSUzI1NiIsImtpZsomethingSomethingsomethingDarkSideOEJBNk
    """

    importer = importer_factory(TEST_CONFIG_WITHOUT_ACCOUNTS)
    assert importer._get_account_for_account_id("any-account-id-1") == "DefaultAccount"



import csv
import logging
import re
from datetime import datetime, timedelta
from decimal import Decimal

import beangulp
from beancount.core import data


class Importer(beangulp.Importer):
    """An importer for PostFinance CSV."""

    def __init__(self, filepattern: str, account: data.Account, currency: str = "CHF"):
        self._filepattern = filepattern
        self._account = account
        self.currency = currency

    def identify(self, filepath: str) -> bool:
        return re.search(self._filepattern, filepath) is not None

    def account(self, filepath: str) -> data.Account:
        return self._account

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        csvfile = open(file=filepath, encoding="windows_1252")
        reader = csv.reader(csvfile, delimiter=";")
        meta = data.new_metadata(filepath, 0)
        entries = []

        for row in reader:
            try:
                book_date_str, text, credit, debit, val_date, balance_str = tuple(row)
                book_date = datetime.strptime(book_date_str, "%Y-%m-%d").date()
                if credit:
                    amount = data.Amount(Decimal(credit), self.currency)
                elif debit:
                    amount = data.Amount(Decimal(debit), self.currency)
                else:
                    amount = None
                if balance_str:
                    balance = data.Amount(Decimal(balance_str), self.currency)
                else:
                    balance = None
            except Exception as e:
                logging.debug(e)
            else:
                logging.debug((book_date, text, amount, val_date, balance))
                posting = data.Posting(self._account, amount, None, None, None, None)
                entry = data.Transaction(
                    meta,
                    book_date,
                    "*",
                    "",
                    text,
                    data.EMPTY_SET,
                    data.EMPTY_SET,
                    [posting],
                )
                entries.append(entry)
                # only add balance on SOM
                book_date = book_date + timedelta(days=1)
                if balance and book_date.day == 1:
                    entry = data.Balance(
                        meta, book_date, self._account, balance, None, None
                    )
                    entries.append(entry)

        csvfile.close()
        entries = data.sorted(entries)
        return entries



import datetime
from os import path
from typing import Any

import beangulp
import yaml
from beancount.core import amount, data
from beancount.core.number import D
from dateutil.relativedelta import relativedelta


class Importer(beangulp.Importer):
    """An importer for Scheduled/Recurring Transactions."""

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("schedule.yaml")

    def account(self, filepath: str) -> data.Account:
        return ""

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)
        self.transactions = config["transactions"]

        result = []
        for trx in config["transactions"]:
            for i in reversed(range(1, 6)):
                date = datetime.date.today() + relativedelta(months=-i, day=31)
                result.append(self.createForDate(trx, date))

        return result

    def createForDate(
        self, trx: dict[str, Any], date: datetime.date
    ) -> data.Transaction:
        postings = []
        for post in trx["postings"]:
            amt = None
            if "amount" in post and "currency" in post:
                amt = amount.Amount(D(post["amount"]), post["currency"])

            postings.append(data.Posting(post["account"], amt, None, None, None, None))
        meta = data.new_metadata("schedule", 0)
        return data.Transaction(
            meta,
            date,
            "*",
            "",
            trx["narration"],
            data.EMPTY_SET,
            data.EMPTY_SET,
            postings,
        )



import re
from datetime import date
from decimal import Decimal
from os import path
from typing import Any

import beangulp
import yaml
from beancount.core import amount, data
from beancount.core.number import D
from ibflex import Types, client, parser
from ibflex.enums import CashAction

from tariochbctools.importers.general.priceLookup import PriceLookup


class Importer(beangulp.Importer):
    """An importer for Interactive Broker using the flex query service."""

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("ibkr.yaml")

    def account(self, filepath: str) -> data.Account:
        return ""

    def matches(
        self, trx: Types.CashTransaction, t: Any, account: data.Account
    ) -> bool:
        p = re.compile(r".* (?P<perShare>\d+\.?\d+) PER SHARE")

        trxPerShareGroups = p.search(trx.description)
        tPerShareGroups = p.search(t["description"])

        trxPerShare = trxPerShareGroups.group("perShare") if trxPerShareGroups else ""
        tPerShare = tPerShareGroups.group("perShare") if tPerShareGroups else ""

        return (
            t["date"] == trx.dateTime
            and t["symbol"] == trx.symbol
            and trxPerShare == tPerShare
            and t["account"] == account
        )

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)
        token = config["token"]
        queryId = config["queryId"]

        priceLookup = PriceLookup(existing, config["baseCcy"])

        response = client.download(token, queryId)
        statement = parser.parse(response)
        assert isinstance(statement, Types.FlexQueryResponse)

        result = []
        for stmt in statement.FlexStatements:
            transactions: list = []
            account = stmt.accountId
            for trx in stmt.Trades:
                result.append(
                    self.createBuy(
                        trx.tradeDate,
                        account,
                        trx.symbol.rstrip("z"),
                        trx.quantity,
                        trx.currency,
                        trx.tradePrice,
                        amount.Amount(
                            round(-trx.ibCommission, 2), trx.ibCommissionCurrency
                        ),
                        amount.Amount(round(trx.netCash, 2), trx.currency),
                        config["baseCcy"],
                        trx.fxRateToBase,
                    )
                )

            for trx in stmt.CashTransactions:
                existingEntry = None
                if CashAction.DIVIDEND == trx.type or CashAction.WHTAX == trx.type:
                    existingEntry = next(
                        (
                            t
                            for t in transactions
                            if self.matches(trx, t, stmt.accountId)
                        ),
                        None,
                    )

                if existingEntry:
                    if CashAction.WHTAX == trx.type:
                        existingEntry["whAmount"] += trx.amount
                    else:
                        existingEntry["amount"] += trx.amount
                        existingEntry["description"] = trx.description
                        existingEntry["type"] = trx.type
                else:
                    if CashAction.WHTAX == trx.type:
                        amt = 0
                        whAmount = trx.amount
                    else:
                        amt = trx.amount
                        whAmount = 0

                    transactions.append(
                        {
                            "date": trx.dateTime,
                            "symbol": trx.symbol,
                            "currency": trx.currency,
                            "amount": amt,
                            "whAmount": whAmount,
                            "description": trx.description,
                            "type": trx.type,
                            "account": account,
                        }
                    )

            for trx in transactions:
                if trx["type"] == CashAction.DIVIDEND:
                    asset = trx["symbol"].rstrip("z")
                    payDate = trx["date"].date()
                    totalDividend = trx["amount"]
                    totalWithholding = -trx["whAmount"]
                    totalPayout = totalDividend - totalWithholding
                    currency = trx["currency"]
                    account = trx["account"]

                    result.append(
                        self.createDividen(
                            totalPayout,
                            totalWithholding,
                            asset,
                            currency,
                            payDate,
                            priceLookup,
                            trx["description"],
                            account,
                        )
                    )

        return result

    def createDividen(
        self,
        payout: Decimal,
        withholding: Decimal,
        asset: str,
        currency: str,
        date: date,
        priceLookup: PriceLookup,
        description: str,
        account: str,
    ) -> data.Transaction:
        narration = "Dividend: " + description
        liquidityAccount = self.getLiquidityAccount(account, currency)
        incomeAccount = self.getIncomeAccount(account)
        assetAccount = self.getAssetAccount(account, asset)

        price = priceLookup.fetchPrice(currency, date)

        postings = [
            data.Posting(
                assetAccount, amount.Amount(D(0), asset), None, None, None, None
            ),
            data.Posting(
                liquidityAccount,
                amount.Amount(payout, currency),
                None,
                price,
                None,
                None,
            ),
        ]
        if withholding > 0:
            receivableAccount = self.getReceivableAccount(account)
            postings.append(
                data.Posting(
                    receivableAccount,
                    amount.Amount(withholding, currency),
                    None,
                    None,
                    None,
                    None,
                )
            )
        postings.append(data.Posting(incomeAccount, None, None, None, None, None))

        meta = data.new_metadata("dividend", 0, {"account": account})
        return data.Transaction(
            meta, date, "*", "", narration, data.EMPTY_SET, data.EMPTY_SET, postings
        )

    def createBuy(
        self,
        date: date,
        account: data.Account,
        asset: str,
        quantity: Decimal,
        currency: str,
        price: Decimal,
        commission: amount.Amount,
        netCash: amount.Amount,
        baseCcy: str,
        fxRateToBase: Decimal,
    ) -> data.Transaction:
        narration = "Buy"
        feeAccount = self.getFeeAccount(account)
        liquidityAccount = self.getLiquidityAccount(account, currency)
        assetAccount = self.getAssetAccount(account, asset)

        liquidityPrice = None
        if currency != baseCcy:
            price = price * fxRateToBase
            commission = amount.Amount(
                round(commission.number * fxRateToBase, 2), baseCcy
            )
            liquidityPrice = amount.Amount(fxRateToBase, baseCcy)

        postings = [
            data.Posting(
                assetAccount,
                amount.Amount(quantity, asset),
                data.CostSpec(price, None, baseCcy, None, None, False),
                None,
                None,
                None,
            ),
            data.Posting(feeAccount, commission, None, None, None, None),
            data.Posting(
                liquidityAccount,
                netCash,
                None,
                liquidityPrice,
                None,
                None,
            ),
        ]

        meta = data.new_metadata("buy", 0, {"account": account})
        return data.Transaction(
            meta, date, "*", "", narration, data.EMPTY_SET, data.EMPTY_SET, postings
        )

    def getAssetAccount(self, account: str, asset: str) -> data.Account:
        return f"Assets:{account}:Investment:IB:{asset}"

    def getLiquidityAccount(self, account: str, currency: str) -> data.Account:
        return f"Assets:{account}:Liquidity:IB:{currency}"

    def getReceivableAccount(self, account: str) -> data.Account:
        return f"Assets:{account}:Receivable:Verrechnungssteuer"

    def getIncomeAccount(self, account: str) -> data.Account:
        return f"Income:{account}:Interest"

    def getFeeAccount(self, account: str) -> data.Account:
        return f"Expenses:{account}:Fees"



import re
from typing import Any

from tariochbctools.importers.general import mt940importer


def strip_newline(string: str) -> str:
    return string.replace("\n", "").replace("\r", "")


class BCGEImporter(mt940importer.Importer):
    def prepare_payee(self, trxdata: dict[str, Any]) -> str:
        transaction_details = strip_newline(trxdata["transaction_details"])
        payee = re.search(r"ORDP/([^/]+)", transaction_details)
        if payee is None:
            return ""
        else:
            return payee.group(1)

    def prepare_narration(self, trxdata: dict[str, Any]) -> str:
        transaction_details = strip_newline(trxdata["transaction_details"])
        extra_details = strip_newline(trxdata["extra_details"])
        beneficiary = re.search(r"/BENM/([^/]+)", transaction_details)
        remittance = re.search(r"/REMI/([^/]+)", transaction_details)
        narration = []
        if beneficiary is not None:
            narration.append("Beneficiary: %s" % beneficiary.group(1))
        if remittance is not None:
            narration.append("Remittance: %s" % remittance.group(1))
        return "%s - %s" % (extra_details, ",".join(narration))



from datetime import date

from beancount.core import amount, data, prices
from beancount.core.number import D


class PriceLookup:
    def __init__(self, existing: data.Entries, baseCcy: str):
        if existing:
            self.priceMap = prices.build_price_map(existing)
        else:
            self.priceMap = None
        self.baseCcy = baseCcy

    def fetchPriceAmount(self, instrument: str, date: date) -> data.Amount:
        if self.priceMap:
            price = prices.get_price(
                self.priceMap, tuple([instrument, self.baseCcy]), date
            )
            return price[1]
        else:
            return D(1)

    def fetchPrice(self, instrument: str, date: date) -> data.Amount:
        if instrument == self.baseCcy:
            return None

        return amount.Amount(self.fetchPriceAmount(instrument, date), self.baseCcy)



import json
from unittest.mock import MagicMock, call

import pytest
from beancount.core import data
from beancount.core.number import D

from tariochbctools.importers.quickfile import importer as qfimp

# pylint: disable=protected-access

TEST_CONFIG = b"""
account_number: "YOUR_ACCOUNT_NUMBER"
api_key: SOME_API_KEY
app_id: SOME_APP_ID
accounts:
    1200: Assets:Other
    1201: Assets:Savings
transaction_count: 200
"""

# Example from the quickfile API docs
TEST_TRX = b"""
{
    "TransactionDate": "2017-01-04T12:34:56",
    "Reference": "WATERSTONES REF 364 2177333010 BCC",
    "Amount": -4.75,
    "TagStatus": null,
    "TransactionId": 666
}
"""

TEST_META_DATA = b"""
{
    "RecordsetCount": 6853,
    "ReturnCount": 1,
    "BankName": "Current Account",
    "BankType": "CURRENT",
    "AccountNo": "90145741",
    "SortCode": "208246",
    "Currency": "GBP",
    "CurrentBalance": 3164.97
}
"""

TEST_BANK_SEARCH = b"""
{
  "Bank_Search": {
    "Header": {
      "MessageType": "Response",
      "SubmissionNumber": "8ebba60a-95ce-44af-a8c7-e827ac1ae117"
    },
    "Body": {
      "MetaData": {
        "RecordsetCount": 6853,
        "ReturnCount": 1,
        "BankName": "Current Account",
        "BankType": "CURRENT",
        "AccountNo": "90145741",
        "SortCode": "208246",
        "Currency": "GBP",
        "CurrentBalance": 3164.97
      },
      "Transactions": {
        "Transaction": [
          {
            "TransactionDate": "2017-01-04T00:00:00",
            "Reference": "WATERSTONES REF 364 2177333010 BCC",
            "Amount": -4.75,
            "TagStatus": null,
            "TransactionId": 0
          }
        ]
      }
    }
  }
}
"""


@pytest.fixture(name="tmp_config")
def tmp_config_fixture(tmp_path):
    config = tmp_path / "quickfile.yaml"
    config.write_bytes(TEST_CONFIG)
    yield config


@pytest.fixture(name="importer")
def quickfile_importer_fixture(tmp_config):
    importer = qfimp.Importer()
    importer._configure(tmp_config, [])
    yield importer


@pytest.fixture(name="importer_factory")
def quickfile_importer_factory(tmp_path):
    """A quickfile importer factory for
    generating an importer with a custom config
    """

    def _importer_with_config(custom_config):
        config = tmp_path / "quickfile.yaml"
        config.write_bytes(custom_config)
        importer = qfimp.Importer()
        importer._configure(config, [])
        return importer

    yield _importer_with_config


@pytest.fixture(name="tmp_trx")
def tmp_trx_fixture():
    j = json.loads(TEST_TRX)
    yield qfimp.QuickFileTransaction(**j)


@pytest.fixture(name="tmp_metadata")
def tmp_metadata_fixture():
    j = json.loads(TEST_META_DATA)
    metadata = qfimp.QuickFileResponseMetaData(**j)
    yield metadata


def test_identify(importer, tmp_config):
    assert importer.identify(tmp_config)


def test_extract_transaction_simple(importer, tmp_trx, tmp_metadata):
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", tmp_metadata, [tmp_trx], invert_sign=False
    )
    data.sanity_check_types(entries[0])


def test_extract_transaction_invert_sign(importer, tmp_trx, tmp_metadata):
    """Show that sign inversion works"""
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", tmp_metadata, [tmp_trx], invert_sign=True
    )
    assert entries[0].postings[0].units.number == -D(str(tmp_trx.Amount))


def test_extract_transaction_has_quickfile_id(importer, tmp_trx, tmp_metadata):
    """Ensure mandatory IDs are in extracted transactions."""
    entries = importer._extract_transaction(
        tmp_trx, "Assets:Other", tmp_metadata, [tmp_trx], invert_sign=False
    )
    assert "quickfile_id" in entries[0].meta


def test_request_header_auth(importer, tmp_config):
    header = importer.quickfile.request_header()
    auth = header.get("Authentication")

    config = importer.config

    assert auth["AccNumber"] == config["account_number"]
    assert auth["MD5Value"] == importer.quickfile.auth_md5(
        config["account_number"],
        config["api_key"],
        importer.quickfile.submission_number,
    )
    assert auth["ApplicationID"] == config["app_id"]


def test_bank_search():
    response = json.loads(TEST_BANK_SEARCH)
    body = response["Bank_Search"]["Body"]
    bs = qfimp.QuickFileBankSearch(**body)
    assert len(bs.Transactions["Transaction"]) == 1


def test_extract_all_accounts(importer, tmp_config):
    importer._extract_bank_transactions = MagicMock()
    importer.extract(tmp_config)
    calls = [call(1200), call(1201)]
    importer._extract_bank_transactions.assert_has_calls(calls, any_order=True)


def test_bank_search_with_dates():
    account_number = "37823"
    transaction_count = 13
    from_date = "2017-01-01"
    to_date = "2017-01-31"
    under_test = qfimp.QuickFile(account_number, "an_api_key", "an_app_id")
    under_test._post = MagicMock()
    under_test._post.return_value = json.loads(TEST_BANK_SEARCH)
    under_test.bank_search(account_number, transaction_count, from_date, to_date)
    expected_search_parameters = {
        "SearchParameters": {
            "ReturnCount": str(transaction_count),
            "Offset": "0",
            "OrderResultsBy": "TransactionDate",
            "OrderDirection": "DESC",
            "NominalCode": str(account_number),
            "FromDate": from_date,
            "ToDate": to_date,
        }
    }
    under_test._post.assert_called_with("bank/search", expected_search_parameters)



class ReferenceDuplicatesComparator:
    def __init__(self, refs=["ref"]):
        self.refs = refs

    def __call__(self, entry1, entry2):
        entry1Refs = set()
        entry2Refs = set()
        for ref in self.refs:
            if ref in entry1.meta:
                entry1Refs.add(entry1.meta[ref])
            if ref in entry2.meta:
                entry2Refs.add(entry2.meta[ref])

        return entry1Refs & entry2Refs



from datetime import date
from os import path
from typing import Any

import beangulp
import bitstamp.client
import yaml
from beancount.core import amount, data
from beancount.core.number import MISSING, D
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from tariochbctools.importers.general.deduplication import ReferenceDuplicatesComparator
from tariochbctools.importers.general.priceLookup import PriceLookup


class Importer(beangulp.Importer):
    """An importer for Bitstamp."""

    def identify(self, filepath: str) -> bool:
        return path.basename(filepath).endswith("bitstamp.yaml")

    def account(self, filepath: str) -> data.Account:
        return ""

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        self.priceLookup = PriceLookup(existing, "CHF")

        with open(filepath) as file:
            config = yaml.safe_load(file)
        self.config = config
        self.client = bitstamp.client.Trading(
            username=config["username"], key=config["key"], secret=config["secret"]
        )
        self.currencies = config["currencies"]
        self._account = config["account"]
        self.otherExpensesAccount = config["otherExpensesAccount"]
        self.capGainAccount = config["capGainAccount"]

        dateCutoff = date.today() + relativedelta(months=-config["monthCutoff"])

        trxs = self.client.user_transactions()
        trxs.reverse()
        result = []
        for trx in trxs:
            entry = self.fetchSingle(trx)
            if entry.date > dateCutoff:
                result.append(entry)

        return result

    def fetchSingle(self, trx: dict[str, Any]) -> data.Transaction:
        id = int(trx["id"])
        type = int(trx["type"])
        date = parse(trx["datetime"]).date()

        posAmt = 0
        posCcy = None
        negAmt = 0
        negCcy = None
        for ccy in self.currencies:
            if ccy in trx:
                amt = D(trx[ccy])
                if amt > 0:
                    posAmt = amt
                    posCcy = ccy.upper()
                elif amt < 0:
                    negAmt = amt
                    negCcy = ccy.upper()

        if type == 0:
            narration = "Deposit"
            if posCcy:
                cost = data.CostSpec(
                    self.priceLookup.fetchPriceAmount(posCcy, date),
                    None,
                    "CHF",
                    None,
                    None,
                    False,
                )
            postings = [
                data.Posting(
                    self._account + ":" + posCcy,
                    amount.Amount(posAmt, posCcy),
                    cost,
                    None,
                    None,
                    None,
                ),
            ]
        elif type == 1:
            narration = "Withdrawal"
            postings = [
                data.Posting(
                    self._account + ":" + negCcy,
                    amount.Amount(negAmt, negCcy),
                    None,
                    None,
                    None,
                    None,
                ),
            ]
        elif type == 2:
            fee = D(trx["fee"])
            if posCcy and negCcy and posCcy.lower() + "_" + negCcy.lower() in trx:
                feeCcy = negCcy
                negAmt -= fee
            else:
                feeCcy = posCcy
                posAmt -= fee

            if feeCcy:
                rateFiatCcy = self.priceLookup.fetchPriceAmount(feeCcy, date)
            if feeCcy == posCcy:
                posCcyCost = None
                posCcyPrice = amount.Amount(rateFiatCcy, "CHF")
                negCcyCost = data.CostSpec(MISSING, None, MISSING, None, None, False)
                negCcyPrice = None
            else:
                posCcyCost = data.CostSpec(
                    None, D(-negAmt * rateFiatCcy), "CHF", None, None, False
                )
                posCcyPrice = None
                negCcyCost = None
                negCcyPrice = amount.Amount(rateFiatCcy, "CHF")

            narration = "Trade"

            postings = [
                data.Posting(
                    self._account + ":" + posCcy,
                    amount.Amount(posAmt, posCcy),
                    posCcyCost,
                    posCcyPrice,
                    None,
                    None,
                ),
                data.Posting(
                    self._account + ":" + negCcy,
                    amount.Amount(negAmt, negCcy),
                    negCcyCost,
                    negCcyPrice,
                    None,
                    None,
                ),
            ]
            if float(fee) > 0:
                postings.append(
                    data.Posting(
                        self.otherExpensesAccount,
                        amount.Amount(round(fee * rateFiatCcy, 2), "CHF"),
                        None,
                        None,
                        None,
                        None,
                    )
                )
            postings.append(
                data.Posting(self.capGainAccount, None, None, None, None, None)
            )

        else:
            raise ValueError("Transaction type " + str(type) + " is not handled")

        meta = data.new_metadata("bitstamp", id, {"ref": str(id)})
        return data.Transaction(
            meta, date, "*", "", narration, data.EMPTY_SET, data.EMPTY_SET, postings
        )

    cmp = ReferenceDuplicatesComparator()
